from abc import ABC, abstractmethod
from typing import Dict, Any, List
from dataclasses import dataclass

from intents import Intent
from intents.model.intent import IntentParameterMetadata
from intents.model.entity import SystemEntityMixin, _EntityMetaclass

class EntityMapping(ABC):
    """
    An Entity Mapping is a (de-)serializer for predicted Entities.

    Most of the times a Mapping is not needed as the Entity can be mapped
    directly to its type (e.g. "3" -> `Number(3)`). However, prediction services
    such as Dialogflow may define system entities of structured types; a notable
    example is Dialogflow's *sys.person* entity, which is returned as `{"name":
    "John"}` and thus needs to be mapped to `Person("John")`.

    Another notable scenario is Date/Time objects. A mapping can be used to
    convert time strings from the Service format to python objects.
    """

    @property
    @abstractmethod
    def entity_cls(self) -> _EntityMetaclass:
        """
        This is the internal entity type that is being mapped.

        >>> mapping = StringEntityMapping(Sys.Any, 'sys.any')
        >>> mapping.entity_cls
        Sys.Any
        """

    @property
    @abstractmethod
    def service_name(self) -> str:
        """
        This is the name of the Entity in the Prediction Service domain.

        >>> mapping = StringEntityMapping(Sys.Any, 'sys.any')
        >>> mapping.service_name
        'sys.any'
        """

    @abstractmethod
    def from_service(self, service_data: Any) -> SystemEntityMixin:
        """
        Converts the Service representation of an Entity (typically the value
        that is returned at prediction time) to an instance of one of the internal Entity
        classes in :class:`intents.model.entities`
        """

    @abstractmethod
    def to_service(self, entity: SystemEntityMixin) -> Any:
        """
        Converts a System Entity instance into a Service representation (typically,
        to be sent as a parameter of a trigger request)

        TODO: this is currently not used (string values are passed straight to
        prediction services).
        """

class StringEntityMapping(EntityMapping):

    entity_cls: _EntityMetaclass = None
    service_name: str = None

    def __init__(self, entity_cls: _EntityMetaclass, service_name: str):
        """
        :param service_name: name of the Entity in the Prediction Service
        """
        self.entity_cls = entity_cls
        self.service_name = service_name

    def from_service(self, service_data: Any) -> SystemEntityMixin:
        return self.entity_cls(service_data)

    def to_service(self, entity: SystemEntityMixin) -> Any:
        return str(entity)

class ServiceEntityMappings(dict):
    """
    Models a list of entity mappings. In addition to a standard list, two things
    are added:

    * Dict-like access to retrieve a mapping given its internal type (e.g.
      `mappings[Sys.Any]`)
    * Consistency check: a mapping list must cover all the entities defined in
      the framework (TODO)
    * Shortcut to define StringEntityMappings like `(Sys.Any, 'sys.any')` (TODO)
    """

    @classmethod
    def from_list(cls, mapping_list: List[EntityMapping]) -> "ServiceEntityMappings":
        result = cls()
        for mapping in mapping_list:
            if mapping in result:
                raise ValueError(f"Mapping {mapping} already defined in list: {mapping_list}")
            result[mapping.entity_cls] = mapping
        return result

@dataclass
class Prediction(ABC):
    """
    This class is meant to abstract prediction results from a generic Prediction
    Service. It uses names from Dialogflow, as it is currently the only
    supported service
    """
    intent_name: str
    confidence: str
    contexts: dict
    parameters_dict: dict
    fulfillment_messages: dict
    fulfillment_text: str = None

    @property
    @abstractmethod
    def entity_mappings(self) -> ServiceEntityMappings:
        """
        A Prediction subclass must know the Entity Mappings of its Prediction
        Service. :meth:`parameters` will use such mappings to convert parameters
        in `parameters_dict` to their generic `Sys.*` type.
        """

    def parameters(self, schema: Dict[str, IntentParameterMetadata]) -> Dict[str, Any]:
        """
        Cast parameters in `Prediction.parameters_dict` according to the given
        schema. This is necessary, because parameter types are not known at
        prediction time, and therefore they need to be determined by the
        corresponding :class:`Intent` class.
        """
        result = {}
        for param_name, param_value in self.parameters_dict.items():
            if param_name not in schema:
                raise ValueError(f"Found parameter {param_name} in Service Prediction, but Intent class does not define it.")
            param_metadata = schema[param_name]
            mapping = self.entity_mappings[param_metadata.entity_cls]
            if param_metadata.is_list:
                if not isinstance(param_value, list):
                    raise ValueError(f"Parameter {param_name} is defined as List, but returned value is not of 'list' type: {param_value}")
                result[param_name] = [mapping.from_service(x) for x in param_value]
            else:
                result[param_name] = mapping.from_service(param_value)
        return result

class PredictionService(ABC):
    """
    This is the interface that Prediction Services must implement to integrate
    with Agents.
    """

    # TODO: use a metaclass to validate this: must be a dict; all Sys entities
    # must be set as keys. Values must be either strings or EntityMapping object
    entity_mappings = None

    _agent: 'intents.Agent'

    @abstractmethod
    def predict_intent(self, message: str) -> Prediction:
        """
        Predict the Intent from the given User message

        :param message: The message to be classified
        """

    @abstractmethod
    def trigger_intent(self, intent: Intent) -> Prediction:
        """
        Trigger the given Intent and return the Service Prediction.
        """



"""
Here we implement :class:`DialogflowPredictionService`, the instance of
:class:`PredictionService` that allows Agents to operate on Dialogflow

Dialogflow makes use of Protobuffer for its data structures, and protobuf may be
tricky to deal with. For instance, `MessageToDict` will convert snake_case to
lowerCamelCase, so API is documented snake_case, protobuf is snake_case,
dialogflow results are camelCase, protobuf converted to dict is camelCase
(unless you use a flag, in that case it could also be snake_case) 💀 This is one
of the reasons this library exists.
"""
import logging
from typing import Union
from dataclasses import dataclass

import google.auth.credentials
from google.cloud.dialogflow_v2.types import TextInput, QueryInput, EventInput
from google.cloud.dialogflow_v2.services.sessions import SessionsClient
from google.cloud.dialogflow_v2.types import DetectIntentResponse
from google.protobuf.json_format import MessageToDict

from intents import Agent, Intent
from intents.prediction_service import PredictionService, Prediction, EntityMapping
from intents.model.intent import TextFulfillmentMessage
from intents.dialogflow_service.auth import resolve_credentials
from intents.dialogflow_service.util import dict_to_protobuf
from intents.dialogflow_service import entities as df_entities

logger = logging.getLogger(__name__)

@dataclass
class DialogflowPrediction(Prediction):
    """
    This is an implementation of :class:`Prediction` that comes from Dialogflow.
    `DialogflowPredictions` are produced by :class:`DialogflowPredictionService`
    """
    df_response: DetectIntentResponse = None

    entity_mappings = df_entities.MAPPINGS


class DialogflowPredictionService(PredictionService):
    """
    This is an implementation of :class:`PredictionService` that enable Agents to
    work as Dialogflow projects
    """

    entity_mappings = df_entities.MAPPINGS

    _agent: Agent

    _credentials: google.auth.credentials.Credentials
    _session_client: SessionsClient
    
    def __init__(self, agent: Agent, google_credentials: Union[str, google.auth.credentials.Credentials]):
        self._agent = agent
        self._credentials = resolve_credentials(google_credentials)
        self._session_client = SessionsClient(credentials=self._credentials)

    @property
    def gcp_project_id(self):
        return self._credentials.project_id

    def predict_intent(self, message: str) -> Prediction:
        text_input = TextInput(text=message, language_code=self._agent.language)
        query_input = QueryInput(text=text_input)
        session_path=self._session_client.session_path(self.gcp_project_id, self._agent._session)
        df_result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = df_result._pb
        return _df_response_to_prediction(df_response)

    def trigger_intent(self, intent: Intent) -> Prediction:
        intent_name = intent.metadata.name
        event_name = Agent._event_name(intent_name)
        event_parameters = {}
        for parameter_name in intent.parameter_schema():
            if parameter_name in intent.__dict__:
                event_parameters[parameter_name] = intent.__dict__[parameter_name]

        logger.info("Triggering event '%s' in session '%s' with parameters: %s", event_name, self._agent._session, event_parameters)
        if not event_parameters:
            event_parameters = {}
            
        event_input = EventInput(
            name=event_name,
            parameters=dict_to_protobuf(event_parameters),
            language_code=self._agent.language
        )
        query_input = QueryInput(event=event_input)
        session_path=self._session_client.session_path(self.gcp_project_id, self._agent._session)
        result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = result._pb

        return _df_response_to_prediction(df_response)

def _df_response_to_prediction(df_response: DetectIntentResponse) -> DialogflowPrediction:
    fulfillment_messages = []
    for message in df_response.query_result.fulfillment_messages:
        fulfillment_messages.append(_build_response_message(message))

    return DialogflowPrediction( # pylint: disable=abstract-class-instantiated
        intent_name=df_response.query_result.intent.display_name,
        parameters_dict=MessageToDict(df_response.query_result.parameters), # TODO: check types
        contexts=[MessageToDict(c) for c in df_response.query_result.output_contexts], # TODO: model
        confidence=df_response.query_result.intent_detection_confidence,
        fulfillment_messages=fulfillment_messages,
        fulfillment_text=df_response.query_result.fulfillment_text,
        df_response=df_response
    )

def _build_response_message(protobuf):
    """
    Build a fulfillment message from a protobuf structure, as it is found in a
    protobuf response (`query_result.fulfillment_messages`)
    """
    if protobuf.HasField('text'):
        return _build_text_fulfillment_message(protobuf)
    else:
        raise ValueError("Unsupported Fulfillment Message: %s", protobuf)

def _build_text_fulfillment_message(protobuf):
    """
    Build a Text Fulfillment Message from the protobuf equivalent of the
    following dict:

    ```
    {
        "text": {
            "text": [
                "Hello, Human"
            ]
        }
    }
    ```

    https://cloud.google.com/dialogflow/es/docs/reference/rpc/google.cloud.dialogflow.v2#text

    We ignore, the fact that `text` is defined as a list: Dialogflow responses
    only have one.
    """
    text_list = protobuf.text.text
    if len(text_list) == 0:
        return TextFulfillmentMessage('')
    if len(text_list) > 1:
        logger.warning('Text Dialogflow response contains more than one text option, only the first will be considered: %s', protobuf)
    return TextFulfillmentMessage(protobuf.text.text[0])



from intents import SessionEntity

class _ContextMetaclass(type):
    """
    This is just used to set a `name` attribute on the Context class
    """

    @property
    def name(cls):
        return cls.__name__

class Context(SessionEntity, metaclass=_ContextMetaclass):
    """
    Context subclasses are generated from the Dialogflow agent. The name of the class corresponds to the name of the context.

    >>> c_greetings.name
    'c_greetings'
    >>> c_greetings().name
    'c_greetings'
    """

    lifespan: int = None

    def __init__(self, lifespan):
        self.lifespan = lifespan

    @property
    def name(self):
        return self.__class__.name



"""
Here we export a :class:`Agent` to Dialogflow format.
"""
import os
import json
import shutil
import logging
import tempfile
import dataclasses
from uuid import uuid1
from typing import List, Dict
from dataclasses import asdict

from intents import Agent
from intents import language
from intents.model.intent import _IntentMetaclass
from intents.model.entity import EntityMixin, SystemEntityMixin, _EntityMetaclass
import intents.dialogflow_service.df_format as df
from intents.dialogflow_service.service import DialogflowPredictionService
from intents.dialogflow_service.entities import MAPPINGS as ENTITY_MAPPINGS

logger = logging.getLogger(__name__)

def export(agent: Agent, output_path: str) -> None:
    """
    Export the given agent to the given path
    """
    assert isinstance(agent, Agent)
    assert isinstance(agent._prediction_service, DialogflowPredictionService)
    agent_cls = agent.__class__

    output_dir = os.path.join(tempfile.gettempdir(), 'dialogflow-agents-export', agent.name)
    intents_dir = os.path.join(output_dir, 'intents')
    entities_dir = os.path.join(output_dir, 'entities')

    if os.path.isdir(intents_dir):
        logger.warning(f"Removing existing intents folder: {intents_dir}")
        shutil.rmtree(intents_dir)

    if os.path.isdir(entities_dir):
        logger.warning(f"Removing existing entities folder: {entities_dir}")
        shutil.rmtree(entities_dir)

    os.makedirs(intents_dir)
    os.makedirs(entities_dir)

    with open(os.path.join(output_dir, 'agent.json'), 'w') as f:
        json.dump(asdict(render_agent(agent)), f, indent=2)

    with open(os.path.join(output_dir, 'package.json'), 'w') as f:
        json.dump({"version": "1.0.0"}, f, indent=2)

    for intent in agent_cls.intents:
        # TODO: handle multiple languages
        language_data = language.intent_language_data(agent_cls, intent)
        rendered_intent = render_intent(intent, language_data)
        with open(os.path.join(intents_dir, f"{intent.metadata.name}.json"), "w") as f:
            json.dump(asdict(rendered_intent), f, indent=2)
        rendered_intent_usersays = render_intent_usersays(agent_cls, intent, language_data.example_utterances)
        with open(os.path.join(intents_dir, f"{intent.metadata.name}_usersays_en.json"), "w") as f:
            usersays_data = [asdict(x) for x in rendered_intent_usersays]
            json.dump(usersays_data, f, indent=2)

    for entity_cls in agent_cls._entities_by_name.values():
        entries = language.entity_language_data(agent_cls, entity_cls)
        rendered_entity = render_entity(entity_cls)
        with open(os.path.join(entities_dir, f"{entity_cls.name}.json"), "w") as f:
            json.dump(asdict(rendered_entity), f, indent=2)
        rendered_entity_entries = render_entity_entries(agent_cls, entries)
        with open(os.path.join(entities_dir, f"{entity_cls.name}_entries_en.json"), "w") as f:
            entries_data = [asdict(x) for x in rendered_entity_entries]
            json.dump(entries_data, f, indent=2)

    if output_path.endswith('.zip'):
        output_path = output_path[:-4]
    shutil.make_archive(output_path, 'zip', output_dir)

#
# Agent
#

def render_agent(agent: Agent):
    google_assistant = df.AgentGoogleAssistant(
        project=agent._prediction_service.gcp_project_id,
        oAuthLinking=df.AgentGoogleAssistantOauthLinking()
        # TODO: include Google Assistant configuration
    )

    webhook = df.AgentWebhook(
        # TODO: include Webhook configuration
    )

    return df.Agent(
        displayName=agent.name,
        webhook=webhook,
        googleAssistant=google_assistant
    )

#
# Intent
#

def render_intent(intent_cls: _IntentMetaclass, language_data: language.IntentLanguageData):
    response = df.Response(
        affectedContexts=[df.AffectedContext(c.name, c.lifespan) for c in intent_cls.metadata.output_contexts],
        parameters=render_parameters(intent_cls, language_data.slot_filling_prompts),
        messages=render_responses(intent_cls, language_data.responses),
        action=intent_cls.metadata.action
    )

    return df.Intent(
        id=str(uuid1()),
        name=intent_cls.metadata.name,
        contexts=[c.name for c in intent_cls.metadata.input_contexts],
        responses=[response],
        webhookUsed=intent_cls.metadata.intent_webhook_enabled,
        webhookForSlotFilling=intent_cls.metadata.slot_filling_webhook_enabled,
        events=[df.Event(e) for e in intent_cls.metadata.events]
    )

def render_parameters(intent_cls: _IntentMetaclass, slot_filling_prompts: Dict[str, List[str]]):
    result = []
    for param_name, param_metadata in intent_cls.parameter_schema().items():
        entity_cls = param_metadata.entity_cls
        if issubclass(entity_cls, SystemEntityMixin):
            data_type = ENTITY_MAPPINGS[entity_cls].service_name
        else:
            data_type = entity_cls.name

        result.append(df.Parameter(
            id=str(uuid1()),
            name=param_name,
            required=param_metadata.required,
            dataType=f'@{data_type}',
            value=f"${param_name}",
            defaultValue=param_metadata.default if not param_metadata.required else '',
            isList=param_metadata.is_list,
            prompts=[df.Prompt(value=p) for p in slot_filling_prompts.get(param_name, {})]
        ))
    return result

def render_response(response: language.ResponseUtterance):
    if isinstance(response, language.TextResponseUtterance):
        response: language.TextResponseUtterance
        return df.TextResponseMessage(
            speech=response.choices
        )

def render_responses(intent_cls: _IntentMetaclass, responses: List[language.ResponseUtterance]):
    if not responses:
        return [df.ResponseMessage()]

    return [render_response(r) for r in responses]

def render_utterance_chunk(chunk: language.UtteranceChunk):
    if isinstance(chunk, language.TextUtteranceChunk):
        return df.UsersaysTextChunk(text=chunk.text, userDefined=True)

    if isinstance(chunk, language.EntityUtteranceChunk):
        chunk: language.EntityUtteranceChunk
        if issubclass(chunk.entity_cls, SystemEntityMixin):
            meta = ENTITY_MAPPINGS[chunk.entity_cls].service_name
        else:
            meta = chunk.entity_cls.name

        return df.UsersaysEntityChunk(
            text=chunk.parameter_value,
            alias=chunk.parameter_name,
            meta=f'@{meta}',
            userDefined=True
        )

    raise ValueError(f"Unsupported Utterance Chunk Type: {chunk}")

def render_intent_usersays(agent_cls: type, intent: _IntentMetaclass, examples: List[language.ExampleUtterance]):
    result = []
    for e in examples:
        result.append(df.IntentUsersays(
            id=str(uuid1()),
            data=[render_utterance_chunk(c) for c in e.chunks()]
        ))
    return result

#
# Entity
#

def render_entity(entity_cls: _EntityMetaclass) -> df.Entity:
    metadata = entity_cls.metadata
    return df.Entity(
        id=str(uuid1()),
        name=entity_cls.name,
        isRegexp=metadata.regex_entity,
        automatedExpansion=metadata.automated_expansion,
        allowFuzzyExtraction=metadata.fuzzy_matching
    )

def render_entity_entries(entity_cls: _EntityMetaclass, entries: List[language.EntityEntry]) -> List[df.EntityEntry]:
    result = []
    for e in entries:
        result.append(df.EntityEntry(
            value=e.value,
            synonyms=e.synonyms
        ))
    return result

# from example_agent import ExampleAgent
# agent = ExampleAgent('/home/dario/lavoro/dialogflow-agents/_tmp_agents/learning-dialogflow-5827a2d16c34.json')
# export(agent, '/home/dario/lavoro/dialogflow-agents/TMP_AGENT.zip')



class _EventMetaclass(type):
    
    @property
    def name(cls):
        # TODO: handle CamelCase to EVENT_NAME and strip trailing "_EVENT"
        return cls.__name__.upper()

class Event(metaclass=_EventMetaclass):
    """
    This is a base class for an Event

    .. code-block:: python

        class welcome(Event):
            \"\"\"
            This is an external event that is triggered on the Agent 
            by an external source.
            \"\"\"

    Events names are rendered in Dialogflow with their upper case class name.
    The example above will produce a `WELCOME` event in the intent that
    references it.
    """
    pass



from typing import Any
from dataclasses import dataclass

class _EntityMetaclass(type):

    @property
    def name(cls):
        return cls.__name__

    @property
    def metadata(cls) -> "Entity.Meta":
        """
        TODO: consider removing this altogether, a `metadata` field in Entity
        may be enough
        """
        meta = cls.__dict__.get('meta')
        if not meta:
            return Entity.Meta()

        return meta

class EntityMixin(metaclass=_EntityMetaclass):
    """
    This is a mixin class for entities, that adds :meth:`from_df_response`, to
    build the Entity object from the match data in Dialogflow Responses.
    """

    @classmethod
    def from_df_response(cls, match: Any):
        """
        Buid the Entity object from the match data from a Dialogflow Response.
        Specifically, `match` is the :class:`dict` version of
        `queryResult.parameters.<PARAMETER_NAME>`.
        """
        if isinstance(match, dict):
            return cls(**match)

        return cls(match)

class Entity(str, EntityMixin):

    @dataclass
    class Meta:
        """
        This class is used to set user-controlled Entity parameters
        """
        regex_entity: bool = False
        automated_expansion: bool = False
        fuzzy_matching: bool = False

    meta: Meta = None

class SystemEntityMixin(EntityMixin):
    """
    This is used to signal that the entity is one of the Dialogflow default
    entities and doesn't need language resources.
    """

class Sys:

    class Any(str, SystemEntityMixin):
        """
        Matches any non-empty input
        """

    class Integer(int, SystemEntityMixin):
        """
        Matches integers only
        """

    class Person(str, SystemEntityMixin):
        """
        Matches common given names, last names or their combinations.

        Note that in Dialogflow this is returned as an Object (e.g. `{"name":
        "John"}`), while here we define `Person` as a String. The Dialogflow
        module defines proper entity mappings to handle the conversion.
        """

    class MusicArtist(str, SystemEntityMixin):
        """
        Matches the name of a music artist
        """

    class MusicGenre(str, SystemEntityMixin):
        """
        Matches a music genre (rock, pop, reggae, ...)
        """

    # @dataclass
    # class UnitCurrency(SystemEntityMixin):
    #     """
    #     Number + currency name
    #     """
    #     __df_name__ = "sys.unit-currency"

    #     amount: float
    #     currency: str



"""
Utilities to manage the Agent's language resources. An Agent is defined as a
Python package. The package is expected to have a `language` folder at its top
level, containing language resources for intents and entities, in the for of
YAML files.

TODO: expand
"""
import os
import re
import sys
import logging
from typing import List, Dict, Union
from dataclasses import dataclass

import yaml

import intents
from intents.model.intent import _IntentMetaclass
from intents.model.entity import _EntityMetaclass

logger = logging.getLogger(__name__)

def agent_language_folder(agent_cls: type):
    main_agent_package_name = agent_cls.__module__.split('.')[0]
    main_agent_package = sys.modules[main_agent_package_name]
    if '__path__' not in main_agent_package.__dict__:
        # TODO: try workdir or something...
        logger.warning("Agent %s doesn't seem to be defined within a package. Language data will not be loaded.", agent_cls)
        return [], []

    agent_folder = main_agent_package.__path__[0]
    language_folder = os.path.join(agent_folder, 'language')
    if not os.path.isdir(language_folder):
        raise ValueError(f"No language folder found for agent {agent_cls} (expected: {language_folder})")

    return language_folder

#
# Intent Language Data
#

class UtteranceChunk:
    """
    An Example Utterance can be seen as a sequence of Chunks, where each Chunk
    is either a mapped Entity, or a plain text string.
    """

@dataclass
class TextUtteranceChunk(UtteranceChunk):
    """
    An Utterance Chunk that is a static, plain text string.
    """
    text: str

@dataclass
class EntityUtteranceChunk(UtteranceChunk):
    """
    An Utterance Chunk that is a matched entity
    """
    entity_cls: _EntityMetaclass
    parameter_name: str
    parameter_value: str

# TODO: check that parameter_value is one of the entries in custom entities
RE_EXAMPLE_PARAMETERS = re.compile(r"\$(?P<parameter_name>[\w]+)\{(?P<parameter_value>[^\}]+)\}")

class ExampleUtterance(str):
    """
    One of the example Utterances of a given Intent.
    """
    
    # TODO: check for escape characters
    def __init__(self, example: str, intent: intents.Intent):
        self._intent = intent
        self.chunks() # Will check parameters
    
    def __new__(cls, example: str, intent: intents.Intent):
        return super().__new__(cls, example)

    def chunks(self):
        """
        Return the Utterance as a sequence of :class:`UtteranceChunk`. Each
        chunk is either a plain text string, or a mapped Entity.

        >>> utterance = ExampleUtterance("My name is $user_name{Guido}!", intents.user_gives_name)
        >>> utterance.chunks()
        [
            TextUtteranceChunk(text="My name is "),
            EntityUtteranceChunk(entity_cls=Sys.Person, parameter_name="user_name", parameter_value="Guido"),
            TextUtteranceChunk(text="!")
        ]
        """
        parameter_schema = self._intent.parameter_schema()
        result = []
        last_end = 0
        for m in RE_EXAMPLE_PARAMETERS.finditer(self):
            m_start, m_end = m.span()
            m_groups = m.groupdict()
            if m_start > 0:
                result.append(TextUtteranceChunk(text=self[last_end:m_start]))
            
            if (parameter_name := m_groups['parameter_name']) not in parameter_schema:
                raise ValueError(f"Example '{self}' references parameter ${parameter_name}, but intent {self._intent.metadata.name} does not define such parameter.")
 
            entity_cls = parameter_schema[parameter_name].entity_cls
            result.append(EntityUtteranceChunk(
                entity_cls=entity_cls,
                parameter_name=m_groups['parameter_name'],
                parameter_value=m_groups['parameter_value']
            ))

            last_end = m_end

        last_chunk = TextUtteranceChunk(text=self[last_end:])
        if last_chunk.text:
            result.append(last_chunk)

        return result

class ResponseUtterance:
    """
    One of the Response Utterances of a given Intent.
    """

class TextResponseUtterance(ResponseUtterance):
    """
    A plain text response. The actual response is picked randomly from a pool of
    choices.
    """

    choices: List[str]

    def __init__(self, choices: Union[str, List[str]]):
        if not isinstance(choices, list):
            assert isinstance(choices, str)
            choices = [choices]

        self.choices = choices

@dataclass
class IntentLanguageData:
    """
    Language data for an Intent consists of three resources:

    * Example Utterances
    * Slot Filling Prompts
    * Responses

    Example Utterances are the messages that Agent will be trained on to
    recognize the Intent.

    Responses, intuitively, are the Agent's response messages that will be sent
    to User once the Intent is recognized.

    Slot Filling Promps are used to solve parameters that couldn't be tagged in
    the original message. For instance a `order_pizza` intent may have a
    `pizza_type` parameter. When User asks "I'd like a pizza" we want to fill
    the slot by asking "What type of pizza?". `slot_filling_prompts` will map
    parameters to their prompts: `{"pizza_type": ["What type of pizza?"]}`
    """
    example_utterances: List[ExampleUtterance]
    slot_filling_prompts: Dict[str, List[str]]
    responses: List[ResponseUtterance]

def intent_language_data(agent_cls: type, intent_cls: _IntentMetaclass) -> IntentLanguageData:
    language_folder = agent_language_folder(agent_cls)

    # TODO: support multiple languages
    language_file = os.path.join(language_folder, "intents", f"{intent_cls.metadata.name}__en.yaml")
    if not os.path.isfile(language_file):
        raise ValueError(f"Language file not found for intent '{intent_cls.metadata.name}'. Expected path: {language_file}. Language files are required even if the intent doesn't need language; in this case, use an empty file.")
    
    with open(language_file, 'r') as f:
        language_data = yaml.load(f.read(), Loader=yaml.FullLoader)

    if not language_data:
        return IntentLanguageData([], {}, [])

    examples_data = language_data.get('examples', [])
    responses_data = language_data.get('responses', [])

    examples = [ExampleUtterance(s, intent_cls) for s in examples_data]
    responses = _build_responses(responses_data)

    return IntentLanguageData(
        example_utterances=examples,
        slot_filling_prompts=language_data.get('slot_filling_prompts', {}),
        responses=responses
    )

def _build_responses(responses_data: dict):
    result = []

    platform: str
    responses: List[dict]
    for platform, responses in responses_data.items():
        if platform != 'default':
            raise NotImplementedError(f"Unsupported platform '{platform}'. Currently, only 'default' is supported")
        for r in responses:
            assert len(r) == 1
            for r_type, r_data in r.items():
                if r_type != 'text':
                    raise NotImplementedError(f"Unsupported response type '{r_type}'. Currently, only 'text' is supported")
                result.append(TextResponseUtterance(r_data))

    return result

#
# Entities
#

@dataclass
class EntityEntry:

    value: str
    synonyms: List[str]

def entity_language_data(agent_cls: type, entity_cls: _EntityMetaclass) -> List[EntityEntry]:
    language_folder = agent_language_folder(agent_cls)

    # TODO: support other languages
    language_file = os.path.join(language_folder, "entities", f"{entity_cls.name}__en.yaml")
    if not os.path.isfile(language_file):
        raise ValueError(f"Language file not found for entity '{entity_cls.name}'. Expected path: {language_file}.")
    
    with open(language_file, 'r') as f:
        language_data = yaml.load(f.read(), Loader=yaml.FullLoader)

    if not language_data:
        return []

    entries_data = language_data.get('entries', [])
    if not isinstance(entries_data, dict):
        raise ValueError(f"Invalid Entity language data for entity {entity_cls.name}. Entries data must be a dict. Entries data: {entries_data}")

    entries = []
    for value, synonyms in entries_data.items():
        if not isinstance(synonyms, list):
            raise ValueError(f"Invalid language data for entry {entity_cls.name}. Synonims data must always be lists. Synonims data for '{value}': '{synonyms}'")
        entries.append(EntityEntry(value, synonyms))

    return entries

# from example_agent import ExampleAgent
# from example_agent import smalltalk

# examples, responses = intent_language_data(ExampleAgent, smalltalk.user_name_give)
# for e in examples:
#     print(e.chunks())

# from example_agent import ExampleAgent
# from example_agent.restaurant import PizzaType

# entity_language_data(ExampleAgent, PizzaType)



from typing import Union

import google.oauth2.service_account
from google.auth.credentials import Credentials


def resolve_credentials(credentials_or_path: Union[str, Credentials]) -> Credentials:
    """
    Determine whether the given object is a valid instance of
    :class:`Credentials`, or a path to a Service Account JSON. In the latter
    case, the JSON is read into an instance of
    :class:`google.oauth2.service_account.Credetials`
    (https://google-auth.readthedocs.io/en/latest/reference/google.oauth2.service_account.html)
    """
    if isinstance(credentials_or_path, Credentials):
        return credentials_or_path

    elif isinstance(credentials_or_path, str):
        return google.oauth2.service_account.Credentials.from_service_account_file(credentials_or_path)
        
    else:
        raise ValueError("Unsupported Dialogflow Credentials type. Please pass the path to a Service Account JSON, or an instance of google.auth.credentials.Credentials")



import logging
import dataclasses
from dataclasses import dataclass, field
from typing import List, Dict, Any, _GenericAlias

from intents.model import context, event, entity

logger = logging.getLogger(__name__)

#
# Fulfillment Messages
#

class FulfillmentMessage:
    """
    This is a base class for Intent Fulfillment Messages
    """

class TextFulfillmentMessage(str, FulfillmentMessage):
    """
    This is a simple text fulfillment message.
    """

#
# Intent
#

@dataclass
class IntentMetadata:
    name: str
    input_contexts: List[context._ContextMetaclass] # TODO: model
    output_contexts: List[context.Context] # TODO: model
    events: List[str]
    action: str = None
    end_of_conversation: bool = False
    intent_webhook_enabled: bool = False
    slot_filling_webhook_enabled: bool = False

class _IntentMetaclass(type):
    # @property
    # def metadata(cls) -> IntentMetadata:
    #     if not cls.metadata:
    #         raise ValueError(f"Intent {cls} has no metadata. You need to register it with @agent.intent before using it")
    #     return cls.metadata
    metadata: IntentMetadata = None

@dataclass
class IntentParameterMetadata:
    name: str
    entity_cls: entity._EntityMetaclass
    is_list: bool
    required: bool
    default: Any

class Intent(metaclass=_IntentMetaclass):
    """
    Represents a predicted intent. This is also used as a base class for the
    intent classes that model a Dialogflow Agent in Python code.

    TODO: check parameter names: no underscore, no reserved names, max length
    """

    @dataclass
    class Meta:
        """
        This is a simpler form of :class:`IntentMetadata`. It is used to specify
        optional metadata when creating Intent classes, without interfering with
        those that are managed by :class:`Agent`.

        TODO: consider merging with IntentMetadata: currently the only
        Agent-managed fields are "name" and "events"
        """
        input_contexts: List[context._ContextMetaclass] = field(default_factory=list)
        output_contexts: List[context.Context] = field(default_factory=list)
        additional_events: List[event.Event] = field(default_factory=list)

    # User fills this with desired extra metadata
    meta: Meta = None

    # :class:`Agent` fills this, integrating Intent.meta with managed fields
    # TODO: make private, with a @property getter
    metadata: IntentMetadata = None
    _df_response = None

    # A :class:`PredictionService` provides this
    prediction: 'intents.Prediction'

    @property
    def confidence(self) -> float:
        return self.prediction.confidence

    @property
    def contexts(self) -> list:
        return self.prediction.contexts

    @property
    def fulfillment_text(self) -> str:
        return self.prediction.fulfillment_text

    def fulfillment_messages(self) -> List[FulfillmentMessage]:
        return self.prediction.fulfillment_messages

    # def fulfillment_messages(self, platform: FulfillmentMessagePlatform=None) -> List[FulfillmentMessage]:
    #     if not platform:
    #         platform = FulfillmentMessagePlatform.PLATFORM_UNSPECIFIED

    #     result_per_platform = defaultdict(list)
    #     for m in self._prediction.fulfillment_messages:
    #         m_platform = FulfillmentMessagePlatform(m.platform)
    #         result_per_platform[m_platform].append(build_response_message(m))

    #     return result_per_platform[platform]

    @classmethod
    def parameter_schema(cls) -> Dict[str, IntentParameterMetadata]:
        """
        Return a dict representing the Intent parameter definition. A key is a
        parameter name, a value is a :class:`IntentParameterMetadata` object.

        TODO: consider computing this in metaclass to cache value and check types
        """
        result = {}
        for param_field in cls.__dict__['__dataclass_fields__'].values():
            # List[...]
            if isinstance(param_field.type, _GenericAlias):
                if param_field.type.__dict__.get('_name') != 'List':
                    raise ValueError(f"Invalid typing '{param_field.type}' for parameter '{param_field.name}'. Only 'List' is supported.")

                if len(param_field.type.__dict__.get('__args__')) != 1:
                    raise ValueError(f"Invalid List modifier '{param_field.type}' for parameter '{param_field.name}'. Must define exactly one inner type (e.g. 'List[Sys.Any]')")
                
                # From here on, check the inner type (e.g. List[Sys.any] -> Sys.Any)
                entity_cls = param_field.type.__dict__.get('__args__')[0]
                is_list = True
            else:
                entity_cls = param_field.type
                is_list = False

            required = isinstance(param_field.default, dataclasses._MISSING_TYPE)
            result[param_field.name] = IntentParameterMetadata(
                name=param_field.name,
                entity_cls=entity_cls,
                is_list=is_list,
                required=required,
                default=param_field.default if not required else '',
            )

        return result

    @classmethod
    def from_prediction(cls, prediction: 'intents.Prediction') -> 'Intent':
        """
        Build an :class:`Intent` class from a :class:`Prediction`. In practice:

        #. Match parameters givent the Intent schema
        #. Instantiate the Intent
        #. Set the `prediction` field on the instantiated Intent.
        """
        try:
            parameters = prediction.parameters(cls.parameter_schema())
        except ValueError as exc:
            raise ValueError(f"Failed to match parameters for Intent class '{cls}'. Prediction: {prediction}") from exc

        result = cls(**parameters)
        result.prediction = prediction
        return result



"""
Here we provide the :class:`Agent` base class. You will subclass it when
defining your own Agent.
"""

import re
import logging
from uuid import uuid1
from inspect import isclass
from typing import List, Dict, Union, _GenericAlias
from dataclasses import dataclass

import google.auth.credentials

from intents.model.intent import Intent, IntentMetadata, _IntentMetaclass
from intents.model.entity import EntityMixin, SystemEntityMixin, _EntityMetaclass
from intents.model.context import Context, _ContextMetaclass
from intents.model.event import _EventMetaclass
from intents.prediction_service import PredictionService, Prediction

logger = logging.getLogger(__name__)

class Agent:
    """
    As the name suggests, Agent is the base class for your Agents
    project.
    """

    intents: List[Intent] = None
    _intents_by_name: Dict[str, Intent] = None
    _intents_by_event: Dict[str, Intent] = None
    _entities_by_name: Dict[str, _EntityMetaclass] = None
    _contexts_by_name: Dict[str, _ContextMetaclass] = None
    _events_by_name: Dict[str, _EventMetaclass] = None

    _prediction_service: PredictionService
    _session: str

    def __init__(self, google_credentials: Union[str, google.auth.credentials.Credentials], session: str=None, language: str="en"):
        if not session:
            session = f"py-{str(uuid1())}"
        
        from intents.dialogflow_service.service import DialogflowPredictionService
        self._prediction_service = DialogflowPredictionService(self, google_credentials)

        self._session = session
        self.language = language

    @classmethod
    def intent(cls, name: str):
        """
        Returns a decorator for Intent subclasses that:

        #. Turns the Intent subclass into a `dataclass`
        #. Registers the intent in the Agent object
        #. Attach metadata to the decorated Intent class
        #. Check language data (examples and responses)

        .. code-block:: python

            from my_agent_project import MyAgent

            @MyAgent.intent('my_test_intent_name')
            class my_test_intent(Intent):
                a_parameter: str
                another_parameter: str

        """
        if not cls.intents or not cls._intents_by_name and not cls._intents_by_event:
            assert not cls.intents and not cls._intents_by_name and not cls._intents_by_event
            assert not cls._contexts_by_name and not cls._events_by_name
            cls.intents = []
            cls._intents_by_name = {}
            cls._intents_by_event = {}
            cls._entities_by_name = {}
            cls._contexts_by_name = {}
            cls._events_by_name = {}

        name_is_valid, reason = _is_valid_intent_name(name)
        if not name_is_valid:
            raise ValueError(f"Invalid name {name}: {reason}")

        if cls._intents_by_name.get(name):
            raise ValueError(f"Another intent exists with name {name}: {cls._intents_by_name[name]}")

        event_name = Agent._event_name(name)
        if conflicting_intent := cls._intents_by_event.get(event_name):
            raise ValueError(f"Intent name {name} is ambiguous and clashes with {conflicting_intent} ('{conflicting_intent.metadata.name}')")

        def _result_decorator(decorated_cls: _IntentMetaclass):
            if not decorated_cls.meta:
                decorated_cls.meta = Intent.Meta()

            for context_cls in decorated_cls.meta.input_contexts:
                cls._register_context(context_cls)

            for context in decorated_cls.meta.output_contexts:
                cls._register_context(context)

            events = [event_name]
            for event_cls in decorated_cls.meta.additional_events:
                cls._register_event(event_cls, decorated_cls)
                events.append(event_cls.name)

            intent_metadata = IntentMetadata(
                name=name,
                input_contexts=decorated_cls.meta.input_contexts,
                output_contexts=decorated_cls.meta.output_contexts,
                events=events
                # TODO: handle other metadata
            )

            result = dataclass(decorated_cls)
            for param_name, param_metadata in result.parameter_schema().items():
                cls._register_entity(param_metadata.entity_cls, param_name, name)
            decorated_cls.metadata = intent_metadata
            cls.intents.append(result)
            cls._intents_by_name[name] = result
            cls._intents_by_event[event_name] = result
            from intents import language
            language.intent_language_data(cls, result) # Checks that language data is existing and consistent
            return result

        return _result_decorator

    @classmethod
    def _register_entity(cls, entity_cls: _EntityMetaclass, parameter_name: str, intent_name: str):
        if not issubclass(entity_cls, EntityMixin):
            raise ValueError(f"Invalid type '{entity_cls}' for parameter '{parameter_name}' in Intent '{intent_name}': must be an Entity. Try 'intents.Sys.Any' if you are unsure.")

        if issubclass(entity_cls, SystemEntityMixin):
            return

        existing_cls = cls._entities_by_name.get(entity_cls.name)
        if not existing_cls:
            from intents import language
            language.entity_language_data(cls, entity_cls) # Checks that language data is existing and consistent
            cls._entities_by_name[entity_cls.name] = entity_cls
            return

        if id(entity_cls) != id(existing_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            entity_cls_path = f"{entity_cls.__module__}.{entity_cls.__qualname__}"
            raise ValueError(f"Two different Entity classes exist with the same name: '{existing_cls_path}' and '{entity_cls_path}'")

    @classmethod
    def _register_context(cls, context_obj_or_cls: Union[_ContextMetaclass, Context]):
        if isinstance(context_obj_or_cls, Context):
            context_cls = context_obj_or_cls.__class__
        elif isclass(context_obj_or_cls) and issubclass(context_obj_or_cls, Context):
            context_cls = context_obj_or_cls
        else:
            raise ValueError(f"Context {context_obj_or_cls} is not a Context instance or subclass")

        existing_cls = cls._contexts_by_name.get(context_cls.name)
        if not existing_cls:
            cls._contexts_by_name[context_cls.name] = context_cls
            return

        if id(context_cls) != id(existing_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            context_cls_path = f"{context_cls.__module__}.{context_cls.__qualname__}"
            raise ValueError(f"Two different Context classes exist with the same name: '{existing_cls_path}' and '{context_cls_path}'")

    @classmethod
    def _register_event(cls, event_cls: _EventMetaclass, intent_cls: _IntentMetaclass):
        existing_cls = cls._events_by_name.get(event_cls.name)

        if not existing_cls:
            cls._events_by_name[event_cls.name] = event_cls
            cls._intents_by_event[event_cls.name] = intent_cls
            return

        if id(existing_cls) != id(event_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            event_cls_path = f"{event_cls.__module__}.{event_cls.__qualname__}"
            raise ValueError(f"Two different Event classes exist with the same name: '{existing_cls_path}' and '{event_cls_path}'")

        # TODO: model different intents with same event and different input
        # context (ok) vs. different intents with same event and same input
        # context (not ok).
        existing_intent = cls._intents_by_event[event_cls.name]
        raise ValueError(f"Event '{event_cls.name}' is alreadt associated to Intent '{existing_intent}'. An Event can only be associated with 1 intent. (differenciation by input contexts is not supported yet)")

    @property
    def name(self):
        return "py-agent" # TODO: parametrize
        # return f"py-{self.gcp_project_id}"

    def predict(self, message: str) -> Intent:
        """
        #. Load persisted session if necessary
        #. Send predict() request, with existing session contexts
        #. Return the right `Intent` subclass

        :param message: The message to be interpreted
        """
        prediction: Prediction = self._prediction_service.predict_intent(message)
        return self._prediction_to_intent(prediction)

    def trigger(self, intent: Intent) -> Intent:
        """
        Trigger the given intent with the given parameters. Return another
        instance of the given Intent, where prediction details have been filled
        in from the response.

        >>> from example_agent import smalltalk
        >>> df_result = agent.trigger(smalltalk.agent_name_give(agent_name='Alice'))
        >>> df_result.confidence
        1.0
        """
        prediction: Prediction = self._prediction_service.trigger_intent(intent)
        return self._prediction_to_intent(prediction)

    def _prediction_to_intent(self, prediction: Prediction) -> Intent:
        """
        Turns a Prediction object into an Intent object
        """
        intent_class = self._intents_by_name.get(prediction.intent_name)
        if not intent_class:
            # TODO: error refers to Dialogflow
            raise ValueError(f"Prediction returned intent '{prediction.intent_name}', but this was not found in Agent definition. Make sure to restore a latest Agent export from `dialogflow_service.export.export()`. If the problem persists, please file a bug on the Dialoglfow Agents repository.")
        return intent_class.from_prediction(prediction)

    def save_session(self):
        """
        Store the current session (most importantly, the list of active
        contexts) to a persisted storage.
        """
        raise NotImplementedError("Context persistence is unsupported yet")

    def load_session(self):
        """
        Load session information (most importantly, the list of active contexts),
        in a format that can be used by :meth:`Agent.predict` to restore the
        state before prediction.
        """
        raise NotImplementedError("Context persistence is unsupported yet")

    @staticmethod
    def _event_name(intent_name: str) -> str:
        """
        Generate the default event name that we associate with every intent.

        >>> _event_name('test.intent_name')
        'E_TEST_INTENT_NAME'

        TODO: This is only used in Dialogflow -> Deprecate and move to DialogflowPredictionService
        """
        return "E_" + intent_name.upper().replace('.', '_')

def _is_valid_intent_name(candidate_name):
    if re.search(r'[^a-zA-Z_\.]', candidate_name):
        return False, "must only contain letters, underscore or dot"

    if candidate_name.startswith('.') or candidate_name.startswith('_'):
        return False, "must start with a letter"

    if "__" in candidate_name:
        return False, "must not contain __"

    return True, None

#
# Example code
#

# from example_agent import ExampleAgent
# from example_agent import smalltalk

# agent = ExampleAgent('/home/dario/lavoro/dialogflow-agents/_tmp_agents/learning-dialogflow-5827a2d16c34.json')
# triggered_intent = agent.trigger(smalltalk.agent_name_give(agent_name='Ugo'))
# predicted_intent = agent.predict("My name is Guido")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['intents', 'intents.dialogflow_service', 'intents.model']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-dialogflow>=2.0.0,<3.0.0', 'pyyaml>=5.4.1,<6.0.0']

setup_kwargs = {
    'name': 'intents',
    'version': '0.0.1',
    'description': 'Define and operate Dialogflow Agents with a simple, code-first, approach',
    'long_description': '# Intents ⛺\n\n**Intents** is a Python library to define and operate Dialogflow Agents with a simple,\ncode-first approach.\n\n> **disclaimer**: this project is not affiliated, associated, authorized,\n> endorsed by, or in any way officially connected with Dialogflow. The names\n> Dialogflow, Google, as well as related names, marks, emblems and images are\n> registered trademarks of their respective owners.\n\n## Project status\n\nThis project is still in early development stage, its API could change without\nnotice. This is a broad overview of the features that are planned and their\ncompletion status.\n\n| Feature           | State  | Note                                                                                |\n|-------------------|--------|-------------------------------------------------------------------------------------|\n| [Agent Definition](STATUS.md#agent-definition)  | 🟡     | Can define basic Intents, with examples, parameters and responses                   |\n| [Cloud Sync](STATUS.md#cloud-sync)        | 🟡     | Can export Agent to a valid Dialogflow ZIP. Cannot yet manage Google Cloud Projects |\n| [Prediction client](STATUS.md#prediction-client) | 🟡     | Can act as a client for predictions and triggers. Cannot receive webhook requests         |\n\nA more detailed view of the single features is reported in [STATUS.md](STATUS.md)\n\n## Install\n\n*Intents* can be installed with PIP as a standard Python package\n\n```sh\npip install intents\n```\n\n## Usage\n\nCheck out the included `example_agent/` to explore *Intents* approach to\nAgent definition. In short, that is a full Agent defined as a set of Python\nclasses (Intents and Entities) and YAML files (language resources).\n\n### Export Agent\n\nA Dialogflow-compatible Agent ZIP can be built as follows:\n\n```py\nfrom example_agent import ExampleAgent\nfrom intents.dialogflow_service.export import export\n\nagent = ExampleAgent(\'/path/to/service_account.json\')\nexport(agent, \'/any/path/ExampleAgent.zip\')\n```\n\n`ExampleAgent.zip` can be loaded into an existing Dialogflow project by using the\nstandard *"Settings > Export and Import > Restore from ZIP"* feature. Also via\nAPI, if you are familiar with it.\n\n### Predict Intent\n\nUploaded Agents can be accessed with a human-friendly API:\n\n```py\nfrom example_agent import ExampleAgent\n\nagent = ExampleAgent(\'/path/to/service_account.json\')\nresult = agent.predict("My name is Guido")\n\nresult                  # user_name_give(user_name="Guido")\nresult.user_name        # "Guido"\nresult.fulfillment_text # "Hi Guido, I\'m Bot"\nresult.confidence       # 0.84\n```\n\n### Trigger Intent\n\nSame goes for triggering intents:\n\n```py\nfrom example_agent import smalltalk\n\nagent = ExampleAgent(\'/path/to/service_account.json\')\nresult = agent.trigger(smalltalk.agent_name_give(agent_name=\'Ugo\'))\n\nresult.fulfillment_text # "Howdy Human, I\'m Ugo"\n```\n\n## Develop\n\n## Setup\n\nDependencies are managed with Poetry\n(https://python-poetry.org/docs/#installation). This is how you setup your\nenvironment:\n\n    poetry install\n\n## Build Documentation\n\nThis project is documented using Sphinx. This is how you build the documentation site:\n\n```sh\ncd docs/\npoetry run make html\n```\n\nDocumentation will be created in the `docs/_build/` folder.\n\n## Test\n\ntbd\n',
    'author': 'Dario',
    'author_email': 'dario.chi@inventati.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dariowho/intents',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from intents.model.entity import Sys
from intents.prediction_service import EntityMapping, StringEntityMapping, ServiceEntityMappings

class PersonEntityMapping(EntityMapping):
    """
    Dialogflow's `sys.person` entity type is returned as an object like
    `{"name": "Guido"}`
    """

    entity_cls = Sys.Person
    service_name = 'sys.person'

    def from_service(self, service_data: dict) -> Sys.Person:
        # Currently `to_service()` is not used in triggers
        if isinstance(service_data, str):
            service_data = {"name": service_data}
        return Sys.Person(service_data['name'])

    def to_service(self, entity: Sys.Person) -> dict:
        return {"name": str(entity)}

MAPPINGS = ServiceEntityMappings.from_list([
    StringEntityMapping(Sys.Any, "sys.any"),
    StringEntityMapping(Sys.Integer, "sys.number-integer"),
    PersonEntityMapping(),
    StringEntityMapping(Sys.MusicArtist, "sys.music-artist"),
    StringEntityMapping(Sys.MusicGenre, "sys.music-genre"),
])



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

__version__ = "0.0.1"

class SessionEntity:

    _session = None

    def to(self, session: str):
        self._session = session
        return self

from intents.model.intent import Intent
from intents.model.entity import Entity, Sys
from intents.model.context import Context
from intents.model.event import Event
from intents.model.agent import Agent
from intents.prediction_service import PredictionService
